""" This code is available under the [CC BY 4.0](https://creativecommons.org/licenses/by/4.0/) license.
You must give appropriate credit, provide a link to the license, and indicate if changes were made. You may do so in any
reasonable manner, but not in any way that suggests the licensor endorses you or your use.

Original code by [Sander Vervoort](sander.vervoort@haw-hamburg.de)
"""
import numpy as np
import scipy.optimize
import matplotlib.pyplot as plt
from scipy.io import loadmat


def mzm_real(
    voltage: float,
    A: float = 1,
    w: float = 2 * np.pi,
    phase_shift: float = 0,
    offset: float = 0,
) -> float:
    """MZM transmission function with compensation for modulation depth.

    Args:
        voltage (float): voltage
        A (float, optional): amplitude (modulation depth). Defaults to 1.
        w (float, optional): angular frequency (pi / half wave voltage). Defaults to 2*np.pi.
        phase_shift (float, optional): phase shift. Defaults to 0.
        offset (float, optional): offset. Defaults to 0.

    Returns:
        float: transmission of the Mach-Zehnder modulator
    """
    return A * np.cos((w * voltage + phase_shift) / 2) ** 2 + offset


def fit_mzm(voltage: np.array, transmission: np.array) -> dict:
    """Fit mzm_real() to the input time sequence, and return fitting parameters

    Args:
        voltage (np.array): Mach-Zehnder voltage
        transmission (np.array): Mach-Zehnder transmission

    Returns:
        dict: dictionary of fitting parameters
            'amp', 'omega', 'phase_shift', 'offset', 'half_wave_voltage' and 'fitfunc'
    """
    v = np.array(voltage)
    y = np.array(transmission)

    popt, pcov = scipy.optimize.curve_fit(mzm_real, v, y, p0=[1.3, 1.8, np.pi, 1.4])
    A, w, phase_shift, offset = popt
    fitfunc = lambda v: mzm_real(
        voltage=v, A=A, w=w, phase_shift=phase_shift, offset=offset
    )  # A * np.sin(w*t + p) + c
    return {
        "amp": A,
        "omega": w,
        "phase_shift": phase_shift,
        "offset": offset,
        "half_wave_voltage": np.pi / w,
        "fitfunc": fitfunc,
        "maxcov": np.max(pcov),
        "popt": popt,
        "pcov": pcov,
    }


def main():
    pdf_export = False
    if pdf_export:
        plt.rcParams.update(
            {
                "text.usetex": True,
                "font.family": "Palatino",
                "figure.figsize": [10 / 2.54, 6 / 2.54],  # 10cm x 6cm
                "savefig.bbox": "tight",  # prevent missing labels
            }
        )

    transmission = loadmat("data/2022-09-20 SV mzm_voltagesweep.mat")[
        "power7V_1"
    ].flatten()
    voltage = np.linspace(0, 7 * 0.2, transmission.size)
    extended_voltage = np.linspace(0, 9 * 0.2, transmission.size)

    res = fit_mzm(voltage, transmission)
    print(
        str(
            "Amplitude={amp}s, "
            "Angular freq.={omega}s, "
            "phase shift={phase_shift}s, "
            "offset={offset}s, "
            "Max. Cov.={maxcov}s, "
            "half wave voltage={half_wave_voltage}s."
        ).format(**res),
    )

    transmission = transmission * (np.abs(res["amp"]) + res["offset"])
    res = fit_mzm(voltage, transmission)
    print(
        str(
            "Amplitude={amp}s, "
            "Angular freq.={omega}s, "
            "phase shift={phase_shift}s, "
            "offset={offset}s, "
            "Max. Cov.={maxcov}s, "
            "half wave voltage={half_wave_voltage}s."
        ).format(**res),
    )
    # Amplitude=-0.6405110396275105, Angular freq.=1.8509597273016019, phase shift=1.5193769757493079, offset=0.7717870575968742, Max. Cov.=0.006881321966279081, half wave voltage=1.697277705413473

    plt.plot(
        extended_voltage,
        res["fitfunc"](extended_voltage) / (np.abs(res["amp"]) + res["offset"]),
        color="C0",
        label=r"Fit ($\cos^2$)",
    )
    plt.plot(
        voltage,
        transmission / (np.abs(res["amp"]) + res["offset"]),
        "C1x",
        label="Measurement",
    )
    x_stem = np.linspace(
        np.abs(res["phase_shift"] - np.pi),
        res["half_wave_voltage"] + res["phase_shift"] - np.pi,
        3,
    )
    y_stem = [
        mzm_real(i, *res["popt"]) / (np.abs(res["amp"]) + res["offset"]) for i in x_stem
    ]
    plt.stem(x_stem, y_stem, "C0")

    plt.legend(loc="best")
    plt.ylabel(r"$\tau'_\mathrm{a}$")
    plt.xlabel(r"$V$ / kV")

    if pdf_export:
        plt.savefig("figures/mzm_halfwavevoltage.pdf")
    plt.show()

    # Determine standard error
    # A, w, p, c
    std_error = np.sqrt(np.diag(res["pcov"]))
    print(f"Half wave voltage = {res['half_wave_voltage']} +- {std_error[1]}")


if __name__ == "__main__":
    main()
