""" This code is available under the [CC BY 4.0](https://creativecommons.org/licenses/by/4.0/) license.
You must give appropriate credit, provide a link to the license, and indicate if changes were made. You may do so in any
reasonable manner, but not in any way that suggests the licensor endorses you or your use.

Original code by [Sander Vervoort](sander.vervoort@haw-hamburg.de)
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.io import loadmat
import mzm


def Sf_meshgrid(
    f: float,
    fm: float = 1,
    phi_mod: np.array = np.linspace(0, 2 * np.pi, 101),
    phi_zero: np.array = np.linspace(0, 2 * np.pi, 101),
) -> np.array:
    """Calculate Sf for frequency f and meshgrid phi_mod and phi_zero

    Args:
        f (float): frequency
        fm (float, optional): modulation frequency. Defaults to 1 Hz.
        phi_mod (np.array, optional):  modulation phase. Defaults to np.linspace(0, 2 * np.pi, 101) .
        phi_zero (np.array, optional): zero phase. Defaults to np.linspace(0, 2 * np.pi, 101).

    Returns:
        np.array: S(f) for the given meshgrid(phi_mod, phi_zero)
    """
    xx, yy = np.meshgrid(phi_mod, phi_zero, sparse=True)
    zz = np.zeros((yy.shape[0], xx.shape[1]))

    for j, xij in enumerate(xx.flatten()):
        for i, yij in enumerate(yy.flatten()):
            zz[i, j] = mzm.calc_Sf(f, fm=fm, phi_mod=xij, phi_zero=yij)

    return zz


def main():
    pdf_export = False
    if pdf_export:
        plt.rcParams.update(
            {
                "text.usetex": True,
                "font.family": "Palatino",
                "figure.figsize": [16 / 2.54, 10 / 2.54],  # 10cm x 6cm
                "savefig.bbox": "tight",  # prevent missing labels
            }
        )

    phi_zero_init = 0.88 * np.pi
    applied_voltage = 7  # V
    wave_voltage = 8.486  # V

    # plot settings
    bbox = dict(edgecolor="none", facecolor="white", alpha=0.7)

    ## fundametal

    # load measurement
    annots = loadmat("data/2022-09-22 SV mzm_fundamental.mat")
    measurement = np.array(annots["fundamental"])

    # create meshgrid with the same dimensions as the measurement
    # x := AC voltage / half wave voltage, y := DC voltage / half wave voltage
    x = np.linspace(0, applied_voltage, measurement.shape[1]) / wave_voltage
    y = np.linspace(0, applied_voltage, measurement.shape[0]) / wave_voltage

    xx, yy = np.meshgrid(x, y)

    # calculate theoretical values
    phi_mod = x * np.pi + 0.001
    phi = y * np.pi + phi_zero_init

    theory = Sf_meshgrid(f=1, phi_mod=phi_mod, phi_zero=phi)[::-1]

    theory_min, theory_max = np.abs(theory).min(), np.abs(theory).max()

    # plot
    fig, ax = plt.subplots(2, 2, sharex=True, sharey=True)

    c = ax[0, 0].pcolormesh(xx, yy, theory, cmap="YlGnBu", vmin=0, vmax=theory_max)
    ax[0, 0].set_title(r"$|S_\mathrm{p,calc}(f_0)|$", y=1.0, pad=-15, bbox=bbox)

    # set the limits of the plot to the limits of the data
    ax[0, 0].axis([x.min(), x.max(), y.min(), y.max()])
    fig.colorbar(c, ax=ax[0, :], location="right", label=r"Norm.\ Amplitude")

    # scale measurement to the same range as the theory
    measurement_scaled = measurement / measurement.max() * theory_max

    c = ax[0, 1].pcolormesh(
        x,
        y,
        measurement_scaled,
        cmap="YlGnBu",
        vmin=np.abs(measurement_scaled).min(),
        vmax=np.abs(measurement_scaled).max(),
    )
    ax[0, 1].set_title(r"$|S_\mathrm{p,meas}(f_0)|$", y=1.0, pad=-15, bbox=bbox)

    # set the limits of the plot to the limits of the data
    ax[0, 1].axis([x.min(), x.max(), y.min(), y.max()])

    print(
        "normalized Mean Absolute Error (nMAE) relative to the analytical "
        "calculation: {:.2f} %".format(
            np.abs(measurement_scaled - theory).mean() / theory.mean() * 100
        )
    )

    ## 1st harmonic

    # load measurement
    annots = loadmat("data/2022-09-22 SV mzm_firstharmonic.mat")
    measurement = np.array(annots["firstharmonic"])

    theory = Sf_meshgrid(f=2, phi_mod=phi_mod, phi_zero=phi)[::-1]

    theory_min, theory_max = np.abs(theory).min(), np.abs(theory).max()

    c = ax[1, 0].pcolormesh(xx, yy, theory, cmap="YlGnBu", vmin=0, vmax=theory_max)
    ax[1, 0].set_title(r"$|S_\mathrm{p,calc}(2f_0)|$", y=1.0, pad=-15, bbox=bbox)

    # set the limits of the plot to the limits of the data
    ax[1, 0].axis([x.min(), x.max(), y.min(), y.max()])
    fig.colorbar(c, ax=ax[1, :], location="right", label=r"Norm.\ Amplitude")

    # scale measurement to the same range as the theory
    measurement_scaled = measurement / measurement.max() * theory_max

    c = ax[1, 1].pcolormesh(
        x,
        y,
        measurement_scaled,
        cmap="YlGnBu",
        vmin=np.abs(measurement_scaled).min(),
        vmax=np.abs(measurement_scaled).max(),
        label="test",
    )
    ax[1, 1].set_title(r"$|S_\mathrm{p,meas}(2f_0)|$", y=1.0, pad=-15, bbox=bbox)

    # set the limits of the plot to the limits of the data
    ax[1, 1].axis([x.min(), x.max(), y.min(), y.max()])

    ax[0, 0].set_ylabel(r"$\phi_0\,/\,\pi$")
    ax[1, 0].set_ylabel(r"$\phi_0\,/\,\pi$")
    ax[1, 0].set_xlabel(r"$\phi_\mathrm{m}\,/\,\pi$")
    ax[1, 1].set_xlabel(r"$\phi_\mathrm{m}\,/\,\pi$")

    if pdf_export:
        plt.savefig("figures/mzm_pcolormesh_subplots.pdf")
    plt.show()

    print(
        "normalized Mean Absolute Error (nMAE) relative to the analytical "
        "calculation: {:.2f} %".format(
            np.abs(measurement_scaled - theory).mean() / theory.mean() * 100
        )
    )


if __name__ == "__main__":
    main()
